import orjson as json
from typing import Dict, Any, Iterable, Optional, Annotated

import typer
from langchain_community.document_loaders.pdf import PyPDFDirectoryLoader

from chroma_dp import CdpProducer, EmbeddableTextResource
from chroma_dp.processor.langchain_utils import convert_lc_doc_to_chroma_resource


class LangchainPyPDFProducer(CdpProducer[EmbeddableTextResource]):
    def __init__(
        self,
        path: str,
        glob: Optional[str] = None,
        recursive: bool = False,
        batch_size: int = 100,
    ) -> None:
        self.path = path
        self.glob = glob
        self.recursive = recursive
        self.batch_size = batch_size

    def produce(
        self, limit: int = -1, offset: int = 0, **kwargs: Dict[str, Any]
    ) -> Iterable[EmbeddableTextResource]:
        loader = PyPDFDirectoryLoader(
            self.path, glob=self.glob, recursive=self.recursive
        )
        docs = loader.load()
        _start = 0
        for _offset in range(_start, len(docs), self.batch_size):
            for doc in docs[_offset : _offset + self.batch_size]:
                yield convert_lc_doc_to_chroma_resource(doc)


def pdf_import(
    path: Annotated[
        str,
        typer.Argument(
            ...,
            help="The path to the directory containing the PDF files.",
        ),
    ],
    glob: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--glob",
            "-g",
            help="The glob pattern to filter files in the directory.",
        ),
    ] = "**/[!.]*.pdf",
    recursive: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--recursive",
            "-r",
            help="A flag whether to recursively search the directory.",
        ),
    ] = False,
    batch_size: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--batch-size",
            "-b",
            help="The batch size to use when processing the PDF files.",
        ),
    ] = 100,
) -> None:
    """Export PDF files from a directory to ChromaDB."""
    producer = LangchainPyPDFProducer(
        path=path, glob=glob, recursive=recursive, batch_size=batch_size
    )
    for doc in producer.produce():
        typer.echo(json.dumps(doc.model_dump()))



import hashlib
import orjson as json
import os
import sys
import uuid
from abc import ABC, abstractmethod
from typing import Any, Iterable, Annotated, Optional

import typer
from jinja2 import Environment
from overrides import EnforceOverrides, override
from ulid import ULID

from chroma_dp import EmbeddableTextResource, CdpProcessor
from chroma_dp.utils import smart_open


class IDStrategy(ABC, EnforceOverrides):
    @abstractmethod
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        pass


class UUIDStrategy(IDStrategy):
    @override
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        return str(uuid.uuid4())


class ULIDStrategy(IDStrategy):
    def __init__(self) -> None:
        self._ulid = ULID()

    @override
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        return str(self._ulid.generate())


class ExprStrategy(IDStrategy):
    def __init__(self, expr: str) -> None:
        self._expr = expr
        self._ulid = ULID()
        self._jinja_env = Environment()

    @override
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        return str(
            self._jinja_env.from_string(self._expr).render(
                uuid=uuid.uuid4, ulid=self._ulid.generate, **doc.model_dump()
            )
        )


class DocHashStrategy(IDStrategy):
    def __init__(self) -> None:
        self._sha256_hash = hashlib.sha256()

    @override
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        if doc.text_chunk is None:
            raise ValueError("Document text chunk is None")
        self._sha256_hash.update(doc.text_chunk.encode("utf-8"))
        return self._sha256_hash.hexdigest()


class RandomHashStrategy(IDStrategy):
    def __init__(self) -> None:
        self._sha256_hash = hashlib.sha256()

    @override
    def generate_id(self, doc: EmbeddableTextResource) -> str:
        self._sha256_hash.update(os.urandom(32))
        return self._sha256_hash.hexdigest()


# --uuid --ulid --expr "{{ metadata.key }}" --doc-hash sha256
class IdStrategyGenerateProcessor(CdpProcessor[EmbeddableTextResource]):
    def __init__(
        self,
        strategy: IDStrategy = UUIDStrategy(),
    ):
        self._strategy = strategy

    def process(
        self, *, documents: Iterable[EmbeddableTextResource], **kwargs: Any
    ) -> Iterable[EmbeddableTextResource]:
        for doc in documents:
            doc.id = self._strategy.generate_id(doc)
            yield doc


def id_process(
    inf: typer.FileText = typer.Argument(sys.stdin),
    file: Optional[str] = typer.Option(None, "--in", help="The Chroma collection."),
    uuid: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--uuid",
            "-u",
            help="Generate a UUID for each document.",
        ),
    ] = None,
    ulid: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--ulid",
            "-U",
            help="Generate a ULID for each document.",
        ),
    ] = None,
    expr: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--expr",
            "-e",
            help="Generate id based on Jinja expression.",
        ),
    ] = None,
    doc_hash: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--doc-hash",
            "-d",
            help="Generate ID based on resources text chunk hash.",
        ),
    ] = None,
    random_hash: Annotated[
        Optional[bool],
        typer.Option(
            ...,
            "--random-hash",
            "-r",
            help="Generate ID based on random hash.",
        ),
    ] = None,
) -> None:
    """Generates IDs for resources."""

    if not any([uuid, ulid, expr, doc_hash, random_hash]):
        typer.echo("Please specify an id generation strategy.")
        raise typer.Exit(code=1)

    if sum([bool(uuid), bool(ulid), bool(expr), bool(doc_hash), bool(random_hash)]) > 1:
        typer.echo("Please specify only one id generation strategy.")
        raise typer.Exit(code=1)
    strategy: Optional[IDStrategy] = None
    if uuid:
        strategy = UUIDStrategy()

    if ulid:
        strategy = ULIDStrategy()

    if expr:
        strategy = ExprStrategy(expr)

    if doc_hash:
        strategy = DocHashStrategy()

    if random_hash:
        strategy = RandomHashStrategy()

    if strategy is None:
        raise ValueError("Cannot find suitable strategy.")
    processor = IdStrategyGenerateProcessor(
        strategy=strategy,
    )

    def process_docs(in_line: str) -> None:
        doc = EmbeddableTextResource(**json.loads(in_line))
        for doc in processor.process(
            documents=[doc],
        ):
            typer.echo(json.dumps(doc.model_dump()))

    with smart_open(file, inf) as file_or_stdin:
        for line in file_or_stdin:
            process_docs(line)



import orjson as json
import sys
import uuid
from concurrent.futures import ThreadPoolExecutor
from typing import Annotated, Optional, List, Dict, Any

import typer
from chromadb import EmbeddingFunction
from chromadb.api.models import Collection

from chroma_dp.utils import smart_open
from chroma_dp.utils.embedding import (
    SupportedEmbeddingFunctions,
    get_embedding_function_for_name,
)
from chroma_dp.utils.chroma import (
    CDPUri,
    get_client_for_uri,
    remap_features,
    DistanceFunction,
)


def add_to_col(
    col: Collection,
    batch: Dict[str, Any],
    upsert: bool = False,
    ef: EmbeddingFunction = None,
) -> None:
    try:
        if ef:
            batch["embeddings"] = ef(batch["documents"])
        if upsert:
            col.upsert(**batch)
        else:
            col.add(**batch)
    except Exception as e:
        raise e


def chroma_import(
    uri: Annotated[str, typer.Argument(help="The Chroma endpoint.")],
    collection: Annotated[
        Optional[str], typer.Option(help="The Chroma collection.")
    ] = None,
    inf: typer.FileText = typer.Argument(
        sys.stdin, help="Stdin input. Requires to pass `-` as the argument."
    ),
    import_file: Optional[str] = typer.Option(
        None, "--in", help="The file to use for the import instead of stdin."
    ),
    create: Annotated[
        bool, typer.Option(help="Create the Chroma collection if it does not exist.")
    ] = False,
    limit: Annotated[int, typer.Option(help="The limit.")] = -1,
    offset: Annotated[int, typer.Option(help="The offset.")] = 0,
    batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
    embedding_function: Optional[SupportedEmbeddingFunctions] = typer.Option(
        None, "--ef", help="The embedding function."
    ),
    upsert: Annotated[bool, typer.Option(help="Upsert documents.")] = False,
    embed_feature: Annotated[
        str, typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Optional[List[str]] = typer.Option(None,
                                                      '-m',
                                                      '--meta-features',
                                                      help="The metadata features to import."),
    id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "text_chunk",
    distance_function: Optional[DistanceFunction] = typer.Option(
        None,
        "--df",
        help="The distance function to use when creating a collection",
    ),
    max_threads: Optional[int] = typer.Option(
        1, "--max-threads", "-t", help="The maximum number of threads."
    ),
) -> None:
    _embedding_function = None
    if embedding_function is not None:
        _embedding_function = get_embedding_function_for_name(embedding_function)
    if uri is None:
        raise ValueError("Please provide a ChromaDP URI.")
    parsed_uri = CDPUri.from_uri(uri)
    client = get_client_for_uri(parsed_uri)
    _collection = parsed_uri.collection or collection
    _batch_size = parsed_uri.batch_size or batch_size
    _offset = parsed_uri.offset or offset
    _limit = parsed_uri.limit or limit
    _upsert = parsed_uri.upsert or upsert
    _create = parsed_uri.create_collection or create
    _batch: Dict[str, Any] = {
        "documents": [],
        "embeddings": [],
        "metadatas": [],
        "ids": [],
    }
    _distance_function = (
        distance_function or parsed_uri.distance_function or DistanceFunction.l2
    )
    if _create:
        chroma_collection = client.get_or_create_collection(
            _collection, metadata={"hnsw:space": _distance_function.value}
        )
    else:
        chroma_collection = client.get_collection(_collection)
    lc_count = 0
    with smart_open(import_file, inf) as file_or_stdin:
        with ThreadPoolExecutor(max_workers=max_threads) as executor:
            for line in file_or_stdin:
                if lc_count < _offset:
                    continue
                if _limit != -1 and lc_count >= _limit:
                    break
                doc = remap_features(
                    json.loads(line),
                    doc_feature,
                    embed_feature,
                    id_feature,
                    meta_features,
                )
                _batch["documents"].append(doc.text_chunk)
                _batch["embeddings"].append(
                    doc.embedding if _embedding_function is None else None
                )  # call EF?
                _batch["metadatas"].append(doc.metadata)
                _batch["ids"].append(doc.id if doc.id else uuid.uuid4())
                if len(_batch["documents"]) >= _batch_size:
                    executor.submit(
                        add_to_col,
                        chroma_collection,
                        _batch,
                        _upsert,
                        _embedding_function,
                    )
                    _batch = {
                        "documents": [],
                        "embeddings": [],
                        "metadatas": [],
                        "ids": [],
                    }
                lc_count += 1
            if len(_batch["documents"]) > 0:
                executor.submit(
                    add_to_col, chroma_collection, _batch, _upsert, _embedding_function
                )



import orjson as json
import sys
from typing import Any, Iterable, Annotated, Optional, List, Union, Dict

import typer
from jinja2 import Template

from chroma_dp import EmbeddableTextResource, CdpProcessor, Metadata
from chroma_dp.utils import smart_open
from chroma_dp.utils.templating import get_jinja_env


_jinja_env = get_jinja_env()


def process_value(value: str) -> Union[bool, float, int, str, Template]:
    """Parse the value as follows: bool>float>int>string"""
    if value.startswith("'") and value.endswith("'"):
        value = value[1:-1]
    if value.startswith("{{") and value.endswith("}}"):
        return _jinja_env.from_string(value)
    if value.lower() == "true":
        return True
    elif value.lower() == "false":
        return False
    elif "." in value:
        try:
            return float(value)
        except ValueError:
            pass
    else:
        try:
            return int(value)
        except ValueError:
            pass
    return value


TemplateMetadata = Dict[str, Union[str, int, float, bool, Template]]


class MetadataProcessor(CdpProcessor[EmbeddableTextResource]):
    def __init__(
        self,
        metadata: Optional[TemplateMetadata] = None,
        remove_keys: Optional[List[str]] = None,
        overwrite: bool = False,
    ):
        self._metadata = metadata
        self._remove_keys = remove_keys
        self._overwrite = overwrite

    def process(
        self, *, documents: Iterable[EmbeddableTextResource], **kwargs: Any
    ) -> Iterable[EmbeddableTextResource]:
        for doc in documents:
            if self._remove_keys and doc.metadata:
                for k in self._remove_keys:
                    if k in doc.metadata.keys():
                        del doc.metadata[k]
            if self._metadata:
                if not doc.metadata:
                    doc.metadata = {}
                for k, v in self._metadata.items():
                    if isinstance(v, Template):
                        # process rendered value
                        doc.metadata[k] = process_value(v.render(**doc.model_dump()))
                    else:
                        if self._overwrite or k not in doc.metadata.keys():
                            doc.metadata[k] = v
            yield doc


def meta_process(
    inf: typer.FileText = typer.Argument(sys.stdin),
    file: Optional[str] = typer.Option(None, "--in", help="The Chroma collection."),
    meta: Annotated[
        Optional[List[str]],
        typer.Option(
            ...,
            "--attr",
            "-a",
            "--meta",
            "-m",
            help="The metadata key value pairs to add e.g. key=value. "
            "The value will be parsed as follows: bool>float>int>string",
        ),
    ] = None,
    remove_keys: Annotated[
        Optional[List[str]],
        typer.Option(
            ...,
            "--remove-key",
            "-k",
            help="The keys to remove from the metadata.",
        ),
    ] = None,
    overwrite: Annotated[
        bool,
        typer.Option(
            ...,
            "--overwrite",
            "-o",
            help="Indicates whether to overwrite the metadata if it already exists. Only applicable for --add.",
        ),
    ] = False,
) -> None:
    """Add or remove metadata."""
    kv_pairs: Metadata = {}
    if not meta and not remove_keys:
        typer.echo(
            "Please specify either --meta or --remove-key",
            err=True,
            color=typer.colors.RED,
            file=sys.stderr,
        )
        raise typer.Abort()
    if meta:
        for opt in meta:
            try:
                key, value = opt.split("=")
                kv_pairs[key] = process_value(value)
            except ValueError:
                typer.echo(
                    f"Invalid metadata: {opt}",
                    err=True,
                    color=typer.colors.RED,
                    file=sys.stderr,
                )
                raise typer.Abort()
    processor = MetadataProcessor(
        metadata=kv_pairs, remove_keys=remove_keys, overwrite=overwrite
    )

    def process_docs(in_line: str) -> None:
        doc = EmbeddableTextResource(**json.loads(in_line))
        for _doc in processor.process(
            documents=[doc],
        ):
            typer.echo(json.dumps(_doc.model_dump()))

    with smart_open(file, inf) as file_or_stdin:
        for line in file_or_stdin:
            process_docs(line)



import orjson as json
import re
import sys
from typing import Optional, Iterable, Any

import typer

from chroma_dp import EmbeddableTextResource, CdpProcessor


def remove_emojis(text: str) -> str:
    emoji_pattern = re.compile(
        "["
        "\U0001F600-\U0001F64F"  # emoticons
        "\U0001F300-\U0001F5FF"  # symbols & pictographs
        "\U0001F680-\U0001F6FF"  # transport & map symbols
        "\U0001F1E0-\U0001F1FF"  # flags (iOS)
        "\U00002702-\U000027B0"
        "\U000024C2-\U0001F251"
        "]+",
        flags=re.UNICODE,
    )
    return emoji_pattern.sub(r"", text)


class EmojiCleanProcessor(CdpProcessor[EmbeddableTextResource]):
    def __init__(self, metadata_clean: Optional[bool] = False):
        self.metadata_clean = metadata_clean

    def process(
        self, *, documents: Iterable[EmbeddableTextResource], **kwargs: Any
    ) -> Iterable[EmbeddableTextResource]:
        for doc in documents:
            doc.text_chunk = remove_emojis(doc.text_chunk)
            if self.metadata_clean:
                for k, v in doc.metadata.items():
                    if isinstance(v, str):
                        doc.metadata[k] = remove_emojis(v)
            yield doc


def emoji_clean(
    inf: typer.FileText = typer.Argument(sys.stdin),
    file: Optional[str] = typer.Option(None, "--in", help="The Chroma collection."),
    metadata_clean: Optional[bool] = typer.Option(
        False, "--metadata-clean", "-m", help="Whether to clean the metadata too."
    ),
) -> None:
    """Chunk a document."""
    processor = EmojiCleanProcessor(metadata_clean=metadata_clean)

    def process_docs(line: str) -> None:
        doc = EmbeddableTextResource(**json.loads(line))
        for doc in processor.process(
            documents=[doc],
        ):
            typer.echo(json.dumps(doc.model_dump()))

    if file:
        with open(file, "r") as inf:
            for line in inf:
                process_docs(line)
    else:
        for line in inf:
            process_docs(line)



import orjson as json

import sys
from typing import Annotated, Optional, List, Generator, Union, Sequence, Any, Dict
from urllib.parse import urlparse, parse_qs

import datasets
import typer
from datasets import load_dataset, Dataset, concatenate_datasets
from datasets.config import HF_ENDPOINT
from huggingface_hub import DatasetCard, HfApi
from pydantic import BaseModel, Field

from chroma_dp import ChromaDocumentSourceGenerator, EmbeddableTextResource
from chroma_dp.huggingface.utils import _infer_hf_type, int_or_none, bool_or_false
from chroma_dp.utils.chroma import remap_features

hf_commands = typer.Typer()


class HFImportRequest(BaseModel):
    dataset: Union[str, Dataset]
    split: Optional[str] = Field("train", description="The Hugging Face dataset split")
    stream: Optional[bool] = Field(
        False, description="Stream dataset instead of downloading."
    )
    limit: Optional[int] = None
    offset: Optional[int] = None
    document_feature: Optional[str] = Field(..., description="Document feature")
    id_feature: Optional[str] = Field(None, description="ID feature")
    embedding_feature: Optional[str] = Field(None, description="Embedding feature")
    metadata_features: Optional[Sequence[str]] = Field(
        None, description="Metadata features"
    )
    batch_size: Optional[int] = Field(100, description="Batch size")

    class Config:
        arbitrary_types_allowed = True


def _doc_wrapper(
    row: Any,
    document_feature: str,
    embedding_feature: Optional[str],
    id_feature: Optional[str],
    metadata_features: Optional[Sequence[str]],
) -> EmbeddableTextResource:
    doc = EmbeddableTextResource(
        id=row[id_feature] if id_feature else None,
        text_chunk=row[document_feature],
        metadata={k: row[k] for k in metadata_features} if metadata_features else None,
        embedding=row[embedding_feature] if embedding_feature else None,
    )
    return doc


class HFChromaDocumentSourceGenerator(
    ChromaDocumentSourceGenerator[EmbeddableTextResource]
):
    """
    A generator of chroma document from a Hugging Face dataset.
    """

    def __init__(self, import_request: HFImportRequest):
        if isinstance(import_request.dataset, str):
            self._dataset = load_dataset(
                import_request.dataset,
                split=import_request.split,
                streaming=import_request.stream,
            )
        else:
            self._dataset = import_request.dataset

        if (
            not import_request.document_feature
            or import_request.document_feature not in self._dataset.features.keys()
        ):
            raise ValueError(
                f"Document column {import_request.document_feature} not found in dataset"
            )
        self._batch_size = import_request.batch_size or 100
        self._doc_feature = import_request.document_feature
        self._id_feature = import_request.id_feature
        self._embed_feature = import_request.embedding_feature
        self._meta_features = import_request.metadata_features
        self._extract_features = [self._doc_feature]
        if self._id_feature:
            self._extract_features.append(self._id_feature)
        if self._embed_feature:
            self._extract_features.append(self._embed_feature)
        if import_request.metadata_features:
            if not all(
                _metadata_column in self._dataset.features
                for _metadata_column in import_request.metadata_features
            ):
                missing_features = [
                    f
                    for f in import_request.metadata_features
                    if f not in self._dataset.features
                ]
                raise ValueError(
                    f"Metadata feature(s) {missing_features} not found in dataset features {self._dataset.features.keys()}"
                )
            self._extract_features.extend(import_request.metadata_features)

        _dataset_len = (
            self._dataset.num_rows if hasattr(self._dataset, "num_rows") else -1
        )
        if _dataset_len > 0:
            self._limit = (
                min(import_request.limit, _dataset_len)
                if import_request.limit != -1
                else _dataset_len
            )
        else:
            self._limit = import_request.limit or -1
        self._offset = import_request.offset or 0
        self._stream = import_request.stream or False

    def _get_batch(self, offset: int, limit: int) -> Dataset:
        return self._dataset[offset : offset + limit]

    def __iter__(self) -> Generator[EmbeddableTextResource, None, None]:
        if self._stream:
            yield from self._streaming_iterator()
        else:
            end = self._offset + self._limit
            for start in range(self._offset, end, self._batch_size):
                subset = self._dataset[start : min(start + self._batch_size, end)]
                yield from [
                    _doc_wrapper(
                        dict(zip(self._extract_features, values)),
                        self._doc_feature,
                        self._embed_feature,
                        self._id_feature,
                        self._meta_features,
                    )
                    for values in zip(*(subset[key] for key in self._extract_features))
                ]

    def _streaming_iterator(self) -> Generator[EmbeddableTextResource, None, None]:
        count = 0
        for item in self._dataset:
            if count < self._offset:
                continue
            if self._limit is not None and 0 < self._limit <= count:
                break

            yield item
            count += 1


class HFImportUri(BaseModel):
    dataset: Optional[str] = None
    dataset_name: Optional[str] = None
    limit: Optional[int] = None
    offset: Optional[int] = None
    split: Optional[str] = None
    stream: Optional[bool] = None
    id_feature: Optional[str] = None
    doc_feature: Optional[str] = None
    embed_feature: Optional[str] = None
    is_remote: Optional[bool] = None
    meta_features: Optional[List[str]] = None
    private: Optional[bool] = Field(
        False,
        description="Make dataset private on Hugging Face Hub. "
        "Note: This parameter is only applicable to exports.",
    )
    batch_size: Optional[int] = Field(100, description="Batch size")

    @staticmethod
    def from_uri(uri: str) -> "HFImportUri":
        parsed_uri = urlparse(uri)
        query_params = parse_qs(parsed_uri.query)

        if parsed_uri.scheme not in ["file", "hf"]:
            raise ValueError(
                f"Unsupported scheme: {parsed_uri.scheme}. Must be 'hf:` or `file:`."
            )
        dataset = (parsed_uri.hostname or "") + parsed_uri.path
        is_remote = False
        if parsed_uri.scheme == "hf":
            is_remote = True
        dataset_name = parsed_uri.path
        limit = int_or_none(query_params.get("limit", [None])[0])
        offset = int_or_none(query_params.get("offset", [None])[0])
        split = query_params.get("split", [None])[0]
        stream = bool_or_false(query_params.get("stream", [False])[0])
        id_feature = query_params.get("id_feature", [None])[0]
        doc_feature = query_params.get("doc_feature", [None])[0]
        embed_feature = query_params.get("embed_feature", [None])[0]
        meta_features = query_params.get("meta_features", [None])[0]
        private = bool_or_false(query_params.get("private", [False])[0])
        batch_size = int_or_none(query_params.get("batch_size", [100])[0])

        return HFImportUri(
            dataset=dataset,
            dataset_name=dataset_name,
            limit=limit,
            offset=offset,
            split=split,
            stream=stream,
            id_feature=id_feature,
            doc_feature=doc_feature,
            embed_feature=embed_feature,
            meta_features=meta_features.split(",") if meta_features else None,
            private=private,
            is_remote=is_remote,
            batch_size=batch_size,
        )


def hf_import(
    uri: Annotated[
        str, typer.Argument(help="Dataset uri. eg. `hf://user/dataset?split=train`")
    ],
    split: Annotated[
        Optional[str], typer.Option(help="The HuggingFace dataset split")
    ] = "train",
    stream: Annotated[
        bool, typer.Option(help="Stream dataset instead of downloading.")
    ] = False,
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "document",
    embed_feature: Annotated[
        Optional[str], typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    id_feature: Annotated[Optional[str], typer.Option(help="The id feature.")] = "id",
    limit: Annotated[Optional[int], typer.Option(help="The limit.")] = -1,
    offset: Annotated[Optional[int], typer.Option(help="The offset.")] = 0,
    batch_size: Optional[int] = typer.Option(
        100, "--batch-size", "-b", help="The batch size."
    ),
) -> None:
    _hf_uri = HFImportUri.from_uri(uri)
    _dataset = _hf_uri.dataset
    _limit = _hf_uri.limit or limit
    _offset = _hf_uri.offset or offset
    _split = _hf_uri.split or split
    _stream = _hf_uri.stream or stream
    _id_feature = _hf_uri.id_feature or id_feature
    _doc_feature = _hf_uri.doc_feature or doc_feature
    _embed_feature = _hf_uri.embed_feature or embed_feature
    _meta_features = _hf_uri.meta_features or meta_features
    _batch_size = batch_size or _hf_uri.batch_size

    import_request = HFImportRequest(
        dataset=_dataset,
        split=_split,
        stream=_stream,
        limit=_limit,
        offset=_offset,
        document_feature=_doc_feature,
        id_feature=_id_feature,
        embedding_feature=_embed_feature,
        metadata_features=_meta_features,
        batch_size=_batch_size,
    )
    gen = HFChromaDocumentSourceGenerator(import_request)
    for doc in gen:
        typer.echo(json.dumps(doc.model_dump()))


def hf_export(
    uri: Annotated[
        str,
        typer.Argument(
            help="Dataset uri. eg. `hf:user/dataset?split=train` or `file:dataset-name?split=train`"
        ),
    ],
    inf: typer.FileText = typer.Argument(sys.stdin),
    split: Annotated[
        Optional[str], typer.Option(help="The HuggingFace dataset split")
    ] = "train",
    doc_feature: Annotated[
        str, typer.Option(help="The document feature.")
    ] = "text_chunk",
    embed_feature: Annotated[
        Optional[str], typer.Option(help="The embedding feature.")
    ] = "embedding",
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    id_feature: Annotated[str, typer.Option(help="The id feature.")] = "id",
    limit: Annotated[int, typer.Option(help="The limit.")] = -1,
    offset: Annotated[int, typer.Option(help="The offset.")] = 0,
    batch_size: Annotated[int, typer.Option(help="The batch size.")] = 100,
    private: Annotated[
        bool, typer.Option(help="Make dataset private on Hugging Face Hub. ")
    ] = False,
) -> None:
    _hf_uri = HFImportUri.from_uri(uri)
    _dataset = _hf_uri.dataset
    _limit = _hf_uri.limit or limit
    _offset = _hf_uri.offset or offset
    _split = _hf_uri.split or split
    _id_feature = _hf_uri.id_feature or id_feature
    _doc_feature = _hf_uri.doc_feature or doc_feature
    _embed_feature = _hf_uri.embed_feature or embed_feature
    _meta_features = _hf_uri.meta_features or meta_features
    _batch_size = batch_size
    _private = _hf_uri.private or private
    _batch: Dict[str, Any] = {
        "id": [],
        "document": [],
        "embedding": [],
    }

    features = datasets.Features(
        {
            "id": datasets.Value("string"),
            "embedding": datasets.features.Sequence(
                feature=datasets.Value(dtype="float32")
            ),
            "document": datasets.Value("string"),
            # **(metadata_feature if metadata_feature else {})
        }
    )
    features.update()
    dataset = None
    for line in inf:
        doc = remap_features(
            json.loads(line),
            doc_feature,
            embed_feature=embed_feature,
            meta_features=meta_features,
            id_feature=id_feature,
        )

        _batch["id"].append(doc.id)
        _batch["document"].append(doc.text_chunk)
        _batch["embedding"].append(doc.embedding)
        if doc.metadata:
            for key in doc.metadata.keys():
                if f"metadata.{key}" not in features:
                    features[f"metadata.{key}"] = _infer_hf_type(doc.metadata[key])
                _batch[f"metadata.{key}"].append(doc.metadata[key])

        if len(_batch["document"]) >= _batch_size:
            if dataset is None:
                dataset = Dataset.from_dict(
                    _batch,
                    features=features,
                    info=datasets.DatasetInfo(
                        description="Chroma Collection export.", features=features
                    ),
                    split=_split,
                )
            else:
                new_dataset = Dataset.from_dict(
                    _batch,
                    features=features,
                    info=datasets.DatasetInfo(
                        description="Chroma Collection export.", features=features
                    ),
                    split=_split,
                )
                dataset = concatenate_datasets([dataset, new_dataset])
            _batch: Dict[str, Any] = {
                "id": [],
                "document": [],
                "embedding": [],
            }

    if len(_batch["document"]) > 0:
        if dataset is None:
            dataset = Dataset.from_dict(
                _batch,
                features=features,
                info=datasets.DatasetInfo(
                    description="Chroma Collection export.", features=features
                ),
                split=_split,
            )
        else:
            new_dataset = Dataset.from_dict(
                _batch,
                features=features,
                info=datasets.DatasetInfo(
                    description="Chroma Collection export.", features=features
                ),
                split=_split,
            )
            dataset = concatenate_datasets([dataset, new_dataset])
    dataset.save_to_disk("test_dataset")

    if _hf_uri.is_remote:
        dataset.push_to_hub(_hf_uri.dataset, private=_private)
        custom_metadata = {
            "license": "mit",
            "language": "en",
            "pretty_name": f"Chroma export of collection N/A",
            "size_categories": ["n<1K"],
            "x-chroma": {
                "description": "Chroma Dataset",
                "collection": "N/A",
                "metadata": "N/A",
            },
        }
        card = DatasetCard.load(repo_id_or_path=_hf_uri.dataset, repo_type="dataset")
        data_info = card.data
        data_dict = {**data_info.to_dict(), **custom_metadata}
        card.content = f"---\n{str(data_dict)}\n---\n{card.text}"
        HfApi(endpoint=HF_ENDPOINT).upload_file(
            path_or_fileobj=str(card).encode(),
            path_in_repo="README.md",
            repo_id=_hf_uri.dataset,
            repo_type="dataset",
        )



from enum import Enum
from typing import Optional, Dict, Any, List, cast
from urllib.parse import urlparse, parse_qs
import chromadb
from chromadb import ClientAPI, GetResult
from chromadb.api.models.Collection import Collection

from pydantic import BaseModel, Field

from chroma_dp import EmbeddableTextResource


def check_collection_exists(client: ClientAPI, collection_name: str) -> bool:
    """Checks if a collection exists in ChromaDB."""
    collections = client.list_collections()
    return collection_name in [collection.name for collection in collections]


def create_collection(
    client: ClientAPI, collection_name: str, if_not_exist: bool = False
) -> Collection:
    """Creates a collection in ChromaDB."""
    if if_not_exist:
        return client.get_or_create_collection(collection_name)
    return client.create_collection(collection_name)


def get_collection(client: ClientAPI, collection_name: str) -> Collection:
    """Gets a collection in ChromaDB."""
    return client.get_collection(collection_name)


def read_large_data_in_chunks(
    collection: Collection, offset: int = 0, limit: int = 100
) -> GetResult:
    """Reads large data in chunks from ChromaDB."""
    result = collection.get(
        limit=limit, offset=offset, include=["embeddings", "documents", "metadatas"]
    )
    return result


class DistanceFunction(str, Enum):
    l2 = "l2"
    ip = "ip"
    cosine = "cosine"


class CDPUri(BaseModel):
    auth: Optional[Dict[str, str]] = None
    host_or_path: Optional[str] = None
    port: Optional[int] = None
    database: Optional[str] = None
    collection: Optional[str] = None
    tenant: Optional[str] = None
    batch_size: Optional[int] = None
    is_local: Optional[bool] = False
    limit: Optional[int] = Field(
        None,
        description="Limit of documents to export. Note: "
        "This parameter is only valid for chroma exports",
    )
    offset: Optional[int] = None
    create_collection: Optional[bool] = False
    upsert: Optional[bool] = False
    distance_function: Optional[DistanceFunction] = None

    @staticmethod
    def from_uri(uri: str) -> "CDPUri":
        parsed = urlparse(uri)
        if parsed.scheme not in ["http", "https", "file"]:
            raise ValueError(
                f"Unsupported scheme: {parsed.scheme}. Must be 'http', 'https' or 'file'."
            )

        is_local = False
        if parsed.scheme == "file":
            is_local = True
        user_info = parsed.username or None
        password_token = parsed.password or None
        if user_info == "__auth_token__":
            auth = {"type": "token", "token": password_token, "header": "AUTHORIZATION"}
        elif user_info == "__x_chroma_token__":
            auth = {
                "type": "token",
                "token": password_token,
                "header": "X-CHROMA-TOKEN",
            }
        elif user_info is not None:
            auth = {"type": "basic", "username": user_info, "password": password_token}
        else:
            auth = None

        host_or_path = parsed.hostname or ""
        port = parsed.port or "8000"

        # Splitting the path into database and collection
        collection = parsed.path.split("/")[-1]
        if is_local:
            host_or_path = host_or_path + parsed.path[: parsed.path.index(collection)]
        # Parsing query parameters
        query_params = parse_qs(parsed.query)
        database = query_params.get("database", [None])[0]
        tenant = query_params.get("tenant", [None])[0]
        batch_size = query_params.get("batch_size", [None])[0]
        _create_collection = query_params.get("create_collection", [None])[0]
        upsert = query_params.get("upsert", [None])[0]
        limit = query_params.get("limit", [None])[0]
        offset = query_params.get("offset", [None])[0]
        distance_function = query_params.get("df", [None])[0]

        return CDPUri(
            auth=auth,
            host_or_path=host_or_path,
            is_local=is_local,
            port=port,
            database=database,
            collection=collection,
            tenant=tenant,
            batch_size=batch_size,
            create_collection=_create_collection,
            upsert=upsert,
            limit=limit,
            offset=offset,
            distance_function=DistanceFunction[distance_function]
            if distance_function
            else None,
        )


def get_client_for_uri(uri: CDPUri) -> ClientAPI:
    """Gets a ChromaDB client for a given URI."""
    if uri.is_local:
        client = chromadb.PersistentClient(
            path=uri.host_or_path,
            database=uri.database or chromadb.api.DEFAULT_DATABASE,
            tenant=uri.tenant or chromadb.api.DEFAULT_TENANT,
        )
    else:
        client = chromadb.HttpClient(
            host=uri.host_or_path,
            port=f"{uri.port}",
            database=uri.database or chromadb.api.DEFAULT_DATABASE,
            tenant=uri.tenant or chromadb.api.DEFAULT_TENANT,
            # TODO auth
            # settings=chromadb.Settings(
            #     chroma_client_auth_provider="chromadb.auth.token.TokenAuthClientProvider",
            #     chroma_client_auth_credentials=parsed_uri.auth,
            #     chroma_client_auth_token_transport_header="X-CHROMA-TOKEN",
            # )
        )

    return client


def get_default_metadata(in_dict: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Attempts to get default metadata from a dictionary (of EmbeddableTextResource) or returns None."""
    if "metadata" in in_dict:
        return cast(Dict[str, Any], in_dict["metadata"])
    return None


def remap_features(
    in_dict: Dict[str, Any],
    doc_feature: str,
    embed_feature: str,
    id_feature: str,
    meta_features: Optional[List[str]] = None,
) -> EmbeddableTextResource:
    # if standard metadata is present and no specific metadata is provided use the standard metadata
    if "metadata" in in_dict and not meta_features:
        # print("in_dict", in_dict)
        _meta = get_default_metadata(in_dict)
    elif "metadata" in in_dict and meta_features:
        _meta = {k: in_dict["metadata"][k] for k in meta_features}
    elif "metadata" not in in_dict and meta_features:
        _meta = {k: in_dict[k] for k in meta_features}
    else:
        _meta = None
    _doc = in_dict[doc_feature]
    _embed = in_dict[embed_feature] if embed_feature else None
    _id = in_dict[id_feature] if id_feature else None
    return EmbeddableTextResource(
        text_chunk=_doc, embedding=_embed, metadata=_meta, id=_id
    )



import time
from datetime import datetime
from typing import Any

from jinja2 import Environment


def date(date_format: str = "epoch") -> str:
    """Custom filter to format the date based on the input format."""
    if date_format == "epoch":
        return str(time.time())
    else:
        return datetime.now().strftime(date_format)


def now(*args: Any, **kwargs: Any) -> str:
    """Custom filter to return the current time."""
    return str(datetime.now())


def get_jinja_env() -> Environment:
    _env = Environment()
    _env.filters["date"] = date
    _env.filters["now"] = now
    return _env



import orjson as json
from typing import Dict, Any, Iterable, Optional, Annotated, List
import csv
import typer
from langchain_community.document_loaders import CSVLoader
from langchain_core.documents import Document

from chroma_dp import CdpProducer, EmbeddableTextResource
from chroma_dp.processor.langchain_utils import convert_lc_doc_to_chroma_resource


class LangchainCSVProducer(CdpProducer[EmbeddableTextResource]):
    def __init__(
        self,
        path: str,
        document_column: Optional[str] = None,
        delimiter: Optional[str] = ",",
        quotechar: Optional[str] = '"',
        metadata_columns: Optional[List[str]] = None,
        batch_size: Optional[int] = 100,
    ) -> None:
        self.path = path
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.document_column = document_column
        self.metadata_columns = metadata_columns
        self.batch_size = batch_size or 100
        self._csv_args = {
            "delimiter": self.delimiter,
            "quotechar": self.quotechar,
        }
        self._all_columns: Optional[List[str]] = []
        with open(self.path) as f:
            csv_reader = csv.DictReader(f, **self._csv_args)  # type: ignore
            self._all_columns = csv_reader.fieldnames
            if (
                document_column is not None
                and document_column not in csv_reader.fieldnames
            ):
                raise ValueError(
                    f"Document column {document_column} not found in CSV file."
                )
            for mc in metadata_columns or []:
                if mc not in csv_reader.fieldnames:
                    raise ValueError(f"Metadata column {mc} not found in CSV file.")
            if not document_column:
                self._all_columns = self.metadata_columns

    def produce(
        self, limit: int = -1, offset: int = 0, **kwargs: Dict[str, Any]
    ) -> Iterable[EmbeddableTextResource]:
        loader = CSVLoader(
            file_path=self.path,
            metadata_columns=self._all_columns,
            csv_args={"delimiter": self.delimiter, "quotechar": self.quotechar},
        )

        docs = loader.load()
        _start = 0
        for _offset in range(_start, len(docs), self.batch_size):
            for doc in docs[_offset : _offset + self.batch_size]:
                doc_content = doc.page_content
                if self.document_column is not None:
                    doc_content = doc.metadata[self.document_column]
                    del doc.metadata[self.document_column]
                d = Document(page_content=doc_content, metadata=doc.metadata)
                yield convert_lc_doc_to_chroma_resource(d)


def csv_import(
    path: Annotated[
        str,
        typer.Argument(
            ...,
            help="The path to the directory containing the text files.",
        ),
    ],
    doc_feature: Annotated[
        Optional[str], typer.Option(help="The document feature.")
    ] = None,
    meta_features: Annotated[
        Optional[List[str]], typer.Option(help="The metadata features.")
    ] = None,
    batch_size: Annotated[
        Optional[int],
        typer.Option(
            ...,
            "--batch-size",
            "-b",
            help="The batch size to use when processing the text files.",
        ),
    ] = 100,
    delimiter: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--delimiter",
            "-s",
            help="The delimiter to use when parsing the CSV file.",
        ),
    ] = ",",
    quotechar: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--quotechar",
            "-q",
            help="The quotechar to use when parsing the CSV file.",
        ),
    ] = '"',
) -> None:
    """Export text files from a directory to ChromaDB."""
    producer = LangchainCSVProducer(
        path=path,
        document_column=doc_feature,
        metadata_columns=meta_features,
        batch_size=batch_size,
        delimiter=delimiter,
        quotechar=quotechar,
    )
    for doc in producer.produce():
        typer.echo(json.dumps(doc.model_dump()))



import orjson as json
import sys
from typing import Any, Iterable, Annotated, Optional

import typer

from chroma_dp import EmbeddableTextResource, CdpProcessor
from langchain.text_splitter import CharacterTextSplitter

from chroma_dp.processor.langchain_utils import (
    convert_chroma_emb_resource_to_lc_doc,
    convert_lc_doc_to_chroma_resource,
)


class ChunkProcessor(CdpProcessor[EmbeddableTextResource]):
    def __init__(self, type: Optional[str] = "character"):
        self.type = type

    def process(
        self, *, documents: Iterable[EmbeddableTextResource], **kwargs: Any
    ) -> Iterable[EmbeddableTextResource]:
        text_splitter = CharacterTextSplitter(
            separator=kwargs.get("separator") if kwargs.get("separator") else "\n",
            chunk_size=kwargs.get("size"),
            chunk_overlap=kwargs.get("overlap", 0),
            add_start_index=kwargs.get("add_start_index", False),
        )
        for doc in documents:
            split_docs = text_splitter.split_documents(
                [convert_chroma_emb_resource_to_lc_doc(doc)]
            )
            for _, split_doc in enumerate(split_docs):
                yield convert_lc_doc_to_chroma_resource(split_doc, doc.metadata)


def chunk_process(
    size: Annotated[
        int,
        typer.Option(
            ...,
            "--size",
            "-s",
            help="The maximum size of each chunk.",
        ),
    ],
    inf: typer.FileText = typer.Argument(sys.stdin),
    file: Optional[str] = typer.Option(None, "--in", help="The Chroma collection."),
    overlap: Annotated[
        int,
        typer.Option(
            ...,
            "--overlap",
            "-o",
            help="The overlap between chunks",
        ),
    ] = 0,
    separator: Annotated[
        str,
        typer.Option(
            ...,
            "--separator",
            "-p",
            help="The separator character to use for splitting the text.",
        ),
    ] = "\n",
    add_start_index: Annotated[
        bool,
        typer.Option(
            ...,
            "--add-start-index",
            "-a",
            help="A flag whether to add the start index to the metadata",
        ),
    ] = False,
    type: Annotated[
        Optional[str],
        typer.Option(
            ...,
            "--type",
            "-t",
            help="The type of the chunking.",
        ),
    ] = "character",
) -> None:
    """Chunk a document."""
    processor = ChunkProcessor(type=type)

    def process_docs(line: str) -> None:
        doc = EmbeddableTextResource(**json.loads(line))
        for doc in processor.process(
            documents=[doc],
            size=size,
            overlap=overlap,
            add_start_index=add_start_index,
            type=type,
            separator=separator,
        ):
            typer.echo(json.dumps(doc.model_dump()))

    if file:
        with open(file, "r") as inf:
            for line in inf:
                process_docs(line)
    else:
        for line in inf:
            process_docs(line)
